@@ -26,11 +26,11 @@ def get_keywords(str):
     return rkws
 
 
-def get_url_of_doc(doc_id):
-    link = 'http://' + os.environ['SERVER_NAME']
-    if os.environ['SERVER_PORT'] != '80':
+def get_url_of_doc(doc_id, env):
+    link = 'http://' + env['SERVER_NAME']
+    if env['SERVER_PORT'] != '80':
 	link = link + ':80'
-    link = link + os.environ['SCRIPT_NAME']
+    link = link + env['SCRIPT_NAME']
     last = link.rfind('/')
     link = link[:last + 1] + 'GinGin_CGI.py'
     link = link + '/show_id_doc/' + str(doc_id)
@@ -124,8 +124,8 @@ class new_action(object):
     # If doc_id is -1, it is authorized for an action
     # that is not for any document.
     def auth(self, doc_id=-1):
-	user = GINGIN_USER.get_login_user()
-	if not GINGIN_USER.authorize(user, self._act_name, doc_id):
+	user = GINGIN_USER.get_login_user(self.octr)
+	if not GINGIN_USER.authorize(self, user, self._act_name, doc_id):
 	    raise DENY_ACTION(user, self._act_name, doc_id)
 	return user
 
@@ -148,7 +148,8 @@ class new_action(object):
 ## Setup form variables.
 #
 # It sets self.form to be a dictionary for an action.  The dictinary
-# is a respository of variables from user clients.
+# is a respository of variables from user clients.  It also sers
+# self.env to deliver environment variables from HTTP daemons.
 class action_form(object):
     def __init__(self, *args, **kws):
 	super(action_form, self).__init__(*args, **kws)
@@ -195,12 +196,12 @@ class CGI_form(action_form):
     def __full_with_path_info(self, res):
 	encoding = GINGIN_ENCODING
 	try:
-	    if os.environ['HTTP_USER_AGENT'].find(' MSIE ') >= 0:
+	    if self.env['HTTP_USER_AGENT'].find(' MSIE ') >= 0:
 		encoding = 'utf8'
 	except:
 	    pass
 	try:
-	    pinfo = split(os.environ['PATH_INFO'], '/')[2:]
+	    pinfo = split(self.env['PATH_INFO'], '/')[2:]
 	    i = 0
 	    while i < len(pinfo) and i < len(self._ival):
 		var = self._ival[i]
@@ -210,8 +211,9 @@ class CGI_form(action_form):
 	    pass
 	pass
 
-    ## Setup form variables from variables of CGI form and PATH_INFO.
-    def set_form(self, form):
+    ## Setup form/env variables from variables of CGI form and PATH_INFO.
+    def set_form(self, form, env):
+        self.env = env
 	self.form = {}
 	self.__full_with_CGI_form(self.form, form)
 	self.__full_with_path_info(self.form)
@@ -227,8 +229,9 @@ class test_form(action_form):
 	super(test_form, self).__init__(*args, **kws)
 	pass
     
-    def set_form(self, form):
+    def set_form(self, form, env):
 	self.form = form
+        self.env = env
 	pass
     pass
 
@@ -298,22 +301,25 @@ class test_octr(action_octr):
     pass
 
 ## @}
+
 # ========================================
 
+## \defgroup actions_G GinGin's actions.
+# \ingroup app_action_G
+# @{
+
 class default_act(new_action):
     def __init__(self):
         super(default_act, self).__init__('default_act',  ['func'])
         pass
     
     def __call__(self):
-        print '7'
         form = self.form
-        print '8'
         try:
             func = form['func']
         except:
             try:
-                func = split(os.environ['PATH_INFO'], '/')[1]
+                func = split(self.env['PATH_INFO'], '/')[1]
             except:
                 func = ''
                 pass
@@ -331,11 +337,15 @@ class show_id_doc(new_action):
     def __call__(self):
 	from templates import doc
 	from xml_template import escape_url
+
 	doc_id = int(self.form['doc_id'])
+
+        self.auth(doc_id)
 	
-	kws, doc_id, title, doc_body, pdate, uid = GinGin_db.get_id_doc(doc_id)
+	kws, doc_id, title, doc_body, pdate, uid, published = \
+            GinGin_db.get_id_doc(doc_id)
 	content = {}
-	creator = GINGIN_USER.get_user_from_id(uid)
+	creator = GINGIN_USER.get_user_from_id(self, uid)
 	all_kws = GinGin_db.get_all_kws()
 	doc_body = GinGin_db.replace_kws(doc_body, all_kws)
 	
@@ -368,9 +378,12 @@ class show_id_doc_src(new_action):
 	
     def __call__(self):
 	doc_id = int(self.form['doc_id'])
-	#self.auth(doc_id)
-	kws, doc_id, title, doc_body, pdate, uid = GinGin_db.get_id_doc(doc_id)
-	creator = GINGIN_USER.get_user_from_id(uid)
+
+        self.auth(doc_id)
+
+	kws, doc_id, title, doc_body, pdate, uid, published = \
+            GinGin_db.get_id_doc(doc_id)
+	creator = GINGIN_USER.get_user_from_id(self, uid)
 	represent.show_doc_src(kws, doc_id, title, doc_body, pdate, creator)
 	pass
     pass
@@ -385,14 +398,17 @@ class edit_id_doc(new_action):
     def __call__(self):
 	from templates import edit
 	doc_id = int(self.form['doc_id'])
+        self.auth(doc_id)
 	
-	kws, doc_id, title, doc_body, pdate, uid = GinGin_db.get_id_doc(doc_id)
+	kws, doc_id, title, doc_body, pdate, uid, published = \
+            GinGin_db.get_id_doc(doc_id)
 	
 	content = {}
 	content['keywords'] = [{'keyword': kw} for kw in kws]
 	content['doc_id'] = doc_id
 	content['title'] = title
 	content['body'] = doc_body
+        content['published'] = published and 'true' or 'false'
 	
 	edit.str = tools.hack_str
 	
@@ -400,6 +416,26 @@ class edit_id_doc(new_action):
     pass
 
 
+## Show unpublished docs.
+class show_all_unpublished(new_action):
+    def __init__(self):
+        super(show_all_unpublished, self).__init__('show_all_unpublished', [])
+        pass
+
+    def __call__(self):
+        from templates import unpublished
+
+        user = self.auth()
+
+        docs = GinGin_db.get_unpublished_docs()
+        docs = [{'doc_id': doc[0], 'title': doc[1],
+                 'pdate': doc[2], 'creator': doc[3]}
+                for doc in docs]
+        content = {'title': 'List of unpublished documents', 'docs': docs}
+        return content, unpublished.unpublished(self.octr.ofo)
+    pass
+
+
 ## Show all keywords of a GinGin system.
 class show_all_kws(new_action):
     def __init__(self):
@@ -424,10 +460,13 @@ class update(new_action):
     def __init__(self):
 	super(update, self).__init__('update',
                                          ['doc_id', 'title', 'keywords',
-                                          'doc_body', 'update_pdate'])
+                                          'doc_body', 'update_pdate',
+                                          'published'])
         pass
     
     def __call__(self):
+        from tools import get_server_url
+
         form = self.form
 	doc_id = int(form['doc_id'])
 	user = self.auth(doc_id)
@@ -455,12 +494,23 @@ class update(new_action):
 	if update_pdate == 'yes':
 	    update_pdate = True
 	    pass
+
+        try:
+            published = form['published']
+        except:
+            published = None
+            pass
+        if published == 'yes':
+            published = True
+            pass
 	
 	if len(doc_body) > MAX_DOC_SZ:
 	    return represent.error_page('too long', self.octr.ofo)
-        uid = GINGIN_USER.get_id_from_user(user)
-        doc_id = GinGin_db.update_doc(doc_id, title, kws, doc_body, uid, update_pdate)
-        represent.show_update_ok(doc_id, title)
+        uid = GINGIN_USER.get_id_from_user(self, user)
+        doc_id = GinGin_db.update_doc(doc_id, title, kws, doc_body,
+                                      uid, update_pdate, published)
+        base, spath = get_server_url(self.octr.env)
+        represent.show_update_ok(doc_id, title, spath)
         return None
 
 ## Show documents and URLs of a keyword.
@@ -484,7 +534,7 @@ class show_kw_docs(new_action):
 	    return content, keyword.keyword(self.octr.ofo)
 	elif len(docs) == 1:
 	    # why not redirect ?
-	    base, spath = tools.get_server_url()
+	    base, spath = tools.get_server_url(self.env)
 	    tgt = spath + '/show_id_doc/%d'
 	    tools.redirect(self.octr.ofo, tgt % (docs[0][0],))
 	else:
@@ -524,7 +574,7 @@ class add_afile(new_action):
 	afobj = form['afile'].file
 	mtype = form['afile'].type
 	GinGin_db.add_afile_to_doc(doc_id, afkey, afobj, mtype)
-	represent.go_back()
+	represent.go_back(self.env)
         pass
 
 ## Delete an attachment.
@@ -540,7 +590,7 @@ class del_afile(new_action):
 	
 	afkey = form['afkey']
 	GinGin_db.remove_afile_by_afkey(doc_id, afkey)
-	represent.go_back()
+	represent.go_back(self.env)
         pass
 	
 ## Get content of an attachement.
@@ -551,10 +601,11 @@ class get_afile(new_action):
     def __call__(self):
         form = self.form
 	doc_id = int(form['doc_id'])
-	self.auth(doc_id)
+        self.auth(doc_id)
 	
 	afkey = form['afkey']
-	afname, mtype = GinGin_db.get_afname_mtype_by_afkey(doc_id, afkey)
+	afname, mtype = \
+            GinGin_db.get_afname_mtype_by_afkey(doc_id, afkey)
 	if afname != None:
 	    represent.send_file(afname, afkey, mtype)
 	else:
@@ -574,7 +625,7 @@ class show_afiles(new_action):
         
         form = self.form
 	doc_id = int(form['doc_id'])
-	self.auth(doc_id)
+        self.auth(doc_id)
 
 	afkeys = GinGin_db.get_afkeys_doc(doc_id)
         content = {'doc_id': doc_id,
@@ -617,13 +668,17 @@ class del_kw_urls(new_action):
 	self.auth()
 	form = self.form
 	
-	keyword = form['keyword']
-	url_ids = form['url_id']
-	if isinstance(url_ids, types.StringTypes):
-	    url_ids = [url_ids]
-	    pass
-	url_ids = map(lambda x: int(x), url_ids)
-	GinGin_db.del_kw_urls(keyword, url_ids)
+        keyword = form['keyword']
+        if form.has_key('url_id'):
+            url_ids = form['url_id']
+            if isinstance(url_ids, types.StringTypes):
+                url_ids = [url_ids]
+                pass
+            url_ids = map(lambda x: int(x), url_ids)
+            GinGin_db.del_kw_urls(keyword, url_ids)
+        else:
+            url_ids = []
+            pass
 	represent.show_del_kw_urls_ok(keyword, url_ids)
 
 ## Show document links for spider.
@@ -633,7 +688,7 @@ class show_to_robot(new_action):
 	
     def __call__(self):
 	docs = GinGin_db.get_last_doc_ids(GINGIN_TO_ROBOT_N)
-	represent.show_to_robot(docs)
+	represent.show_to_robot(docs, self.env)
 	pass
     pass
 
@@ -686,9 +741,9 @@ class update_url(new_action):
             pass
         except KeyError:
 	    return represent.error_page('paramter error', self.octr.ofo)
-	uid = GINGIN_USER.get_id_from_user(user)
+	uid = GINGIN_USER.get_id_from_user(self, user)
 	url_id = GinGin_db.update_url(url_id, title, url, abstract, uid, kws)
-	base, spath = tools.get_server_url()
+	base, spath = tools.get_server_url(self.env)
 	tools.redirect(self.octr.ofo, '%s/show_kw_urls/%s' % (spath, str(kws[0].decode('utf-8'))))
 	pass
     pass
@@ -760,6 +815,8 @@ class add_comment(new_action):
 	pass
     
     def __call__(self):
+        from tools import get_server_url
+
 	form = self.form
 	
 	doc_id = int(form['doc_id'])
@@ -774,7 +831,8 @@ class add_comment(new_action):
 	msg = form['msg']
 	GinGin_db.add_id_doc_comment(doc_id, name, email, website, msg, key)
 	
-	represent.show_update_ok(doc_id, '')
+        base, spath = get_server_url(self.octr.env)
+        represent.show_update_ok(doc_id, '', spath)
 	pass
     pass
 
@@ -804,14 +862,17 @@ class trackback(new_action):
         form = self.form
 	doc_id = int(form['doc_id'])
 	
+        self.auth(doc_id)
+
 	if form.has_key('__mod'):
 	    __mod = form['__mod']
 	    if __mod in ('rss', 'view'):
-		kws, doc_id, title, doc_body, pdate, uid = GinGin_db.get_id_doc(doc_id)
+		kws, doc_id, title, doc_body, pdate, uid, published = \
+                    GinGin_db.get_id_doc(doc_id)
 		desc = doc_body
 		if len(desc) > GINGIN_TB_BODY_MAX:
 		    desc = desc[:GINGIN_TB_BODY_MAX] + '...'
-		link = get_url_of_doc(doc_id)
+		link = get_url_of_doc(doc_id, self.env)
 		pings = GinGin_db.get_trackback_pings(doc_id)
 		if __mod == 'rss':
                     from templates import tb_rss
@@ -845,7 +906,7 @@ class trackback(new_action):
 	else:
 	    try:
 		if form.has_key('title'):
-		    title = form['title'].value
+		    title = form['title']
 		else:
 		    title = ''
 		    pass
@@ -854,7 +915,7 @@ class trackback(new_action):
 		    pass
 		
 		if form.has_key('excerpt'):
-		    excerpt = form['excerpt'].value
+		    excerpt = form['excerpt']
 		else:
 		    excerpt = ''
 		    pass
@@ -862,13 +923,13 @@ class trackback(new_action):
 		    excerpt = excerpt[:GINGIN_TB_BODY_MAX] + '...'
 		    pass
 		
-		url = form['url'].value
+		url = form['url']
 		if len(url) > 128:
 		    url = url[:128]
 		    pass
 		
 		if form.has_key('blog_name'):
-		    blog_name = form['blog_name'].value
+		    blog_name = form['blog_name']
 		else:
 		    blog_name = ''
 		    pass
@@ -879,7 +940,7 @@ class trackback(new_action):
 	    except KeyError:
                 emsg = 'invalid arguments'
             else:
-                rhost = os.environ['REMOTE_ADDR'] + ':' + os.environ['REMOTE_PORT']
+                rhost = self.env['REMOTE_ADDR'] + ':' + self.env['REMOTE_PORT']
                 if exist_doc_id(doc_id):
                     GinGin_db.trackback_ping(doc_id, title, excerpt, url, blog_name, rhost)
                     emsg = None
@@ -903,6 +964,7 @@ class trackback(new_action):
         pass
     pass
 
+## @}
 
 def full_side_bar_info(out):
     docs = GinGin_db.get_last_docs(GINGIN_MOST_RECENT_N)
@@ -915,58 +977,80 @@ def full_side_bar_info(out):
     return out
 
 
-def GinGin_run_action(form, octr):
+def _get_func_name(form, env):
+    try:
+        func = form['func'].value
+    except KeyError:
+        try:
+            pinfo = env['PATH_INFO']
+        except:
+            return None
+        func = split(pinfo, '/')[1]
+        pass
+    return func
+
+def _fallback_homepage(form, env):
+    func = 'show_id_doc'
+    env = dict(env)
+    env['PATH_INFO'] = '/show_id_doc/1'
+    return func, form, env
+
+def _show_error_page(form, env, octr):
+    import sys
+    import traceback
+    tp, val, tb = sys.exc_info()
+    while tb.tb_next:
+        tb = tb.tb_next
+        pass
+    f = tb.tb_frame
+    lno = traceback.tb_lineno(tb)
+    emsg1 = '%s: %s\n\tin %s, L%d file \'%s\'\n' % (str(tp), str(val),
+                                                    f.f_code.co_name,
+                                                    lno,
+                                                    f.f_code.co_filename)
+    #emsg2 = 'locals: %s\nglobals: %s\n' % (repr(f.f_locals),
+    #                                       repr(f.f_globals))
+    return represent.error_page(emsg1, octr.ofo)
+
+def GinGin_run_action(form, env, octr):
     import mez_xml.tools
     mez_xml.tools.str = tools.hack_str
     
-    try:
-	func = form['func'].value
-    except KeyError:
-	try:
-	    pinfo = split(os.environ['PATH_INFO'], '/')[1:]
-	except KeyError:
-	    mixin(show_id_doc, test_form)
-	    mixin(show_id_doc, CGI_octr)
-	    act = show_id_doc()
-	    act.set_form({'doc_id': '1'})
-	    act.set_octr(octr)
-	    return act()
-	func = pinfo[0]
-	pass
-
-    try:
-	clazz = globals()[func]
-    except KeyError:
-	clazz = default_act
+    func = _get_func_name(form, env)
+    if not func:
+        func, form, env = _fallback_homepage(form, env)
         pass
-	
+    
+    act_clazz = _all_actions_.setdefault(func, default_act)
+    
     try:
-        mixin(clazz, CGI_form)
-        mixin(clazz, CGI_octr)
-        act = clazz()
-        act.set_form(form)
+        act = act_clazz()
+        act.set_form(form, env)
         act.set_octr(octr)
         return act()
     except DENY_ACTION, e:
 	return represent.show_deny(e.user, e.act_name, e.doc_id)
     except user.NO_LOGIN_ERR:
-	GINGIN_USER.show_login()
+	GINGIN_USER.show_login(act)
+    except user.REDIRECT, e:
+        represent.redirect(e.url, e.msg)
     except:
-        import sys
-        import traceback
-        tp, val, tb = sys.exc_info()
-        while tb.tb_next:
-            tb = tb.tb_next
-            pass
-        f = tb.tb_frame
-        lno = traceback.tb_lineno(tb)
-        emsg1 = '%s: %s\n\tin %s, L%d file \'%s\'\n' % (str(tp), str(val),
-                                                     f.f_code.co_name,
-                                                     lno,
-                                                     f.f_code.co_filename)
-        #emsg2 = 'locals: %s\nglobals: %s\n' % (repr(f.f_locals),
-        #                                       repr(f.f_globals))
-        return represent.error_page(emsg1, octr.ofo)
+        return _show_error_page(form, env, octr)
     pass
 
+def _init_actions():
+    global _all_actions_
+    _all_actions_ = {}
+    for name, obj in globals().items():
+        if type(obj) == type and \
+                obj != new_action and \
+                issubclass(obj, new_action):
+            mixin(obj, CGI_form)
+            mixin(obj, CGI_octr)
+            _all_actions_[name] = obj
+            pass
+        pass
+    pass
+
+_init_actions()
 
