@@ -15,7 +15,7 @@ class GG_DB_ERR:
     def __init__(self):
 	pass
 
-def update_doc(doc_id, title, keywords, context, uid, update_pdate):
+def update_doc(doc_id, title, keywords, context, uid, update_pdate, published):
     global cx
     if not cx:
 	cx = sqlite3.connect(GINGIN_DB)
@@ -24,8 +24,8 @@ def update_doc(doc_id, title, keywords, context, uid, update_pdate):
 	#
 	# add a new doc
 	#
-	cmd = 'insert into docs select max(doc_id) + 1, ?, ?, ?, ? from docs'
-	cu.execute(cmd, (title, context, now(), uid))
+	cmd = 'insert into docs(title, doc_body, pdate, creator, published) values(?, ?, ?, ?, ?)'
+	cu.execute(cmd, (title, context, now(), uid, published and 1 or 0))
 	cu.execute('select max(doc_id) from docs')
 	doc_id = int(cu.fetchone()[0])
 	del_set = Set()
@@ -45,11 +45,13 @@ def update_doc(doc_id, title, keywords, context, uid, update_pdate):
 	add_set = new_set - old_set
 	# update
 	if update_pdate:
-	    cmd = 'update docs set title = ?, doc_body = ?, pdate = ?, creator = ? where doc_id = ?'
-	    cu.execute(cmd, (title, context, now(), uid, doc_id))
+	    cmd = 'update docs set title = ?, doc_body = ?, pdate = ?, creator = ?, published = ? where doc_id = ?'
+	    cu.execute(cmd, (title, context, now(), uid,
+                             published and 1 or 0, doc_id))
 	else:
-	    cmd = 'update docs set title = ?, doc_body = ?, creator = ? where doc_id = ?'
-	    cu.execute(cmd, (title, context, uid, doc_id))
+	    cmd = 'update docs set title = ?, doc_body = ?, creator = ?, published = ? where doc_id = ?'
+	    cu.execute(cmd, (title, context, uid,
+                             published and 1 or 0, doc_id))
 	    pass
 	pass
     # remove keywords
@@ -74,7 +76,7 @@ def update_doc(doc_id, title, keywords, context, uid, update_pdate):
 		break
 	    try:
 		# add keywords
-		cu.execute('insert into keywords select max(kw_id) + 1, ? from keywords', (kw,))
+		cu.execute('insert into keywords(keyword) values(?)', (kw,))
 	    except sqlite3.DatabaseError:
 		pass
 	    cu.execute(cmd, (doc_id, kw))
@@ -92,7 +94,7 @@ def get_kw_docs(keyword):
 	cx = sqlite3.connect(GINGIN_DB)
     cu = cx.cursor()
     # get doc with specified keyword
-    cmd = 'select c.doc_id, c.title, c.pdate, c.creator from keywords as a inner join kws_docs as b on a.kw_id = b.kw_id inner join docs as c on b.doc_id = c.doc_id where a.keyword = ?'
+    cmd = 'select c.doc_id, c.title, c.pdate, c.creator from keywords as a inner join kws_docs as b on a.kw_id = b.kw_id inner join docs as c on b.doc_id = c.doc_id where a.keyword = ? and c.published = 1'
     cu.execute(cmd, (keyword,))
     docs = cu.fetchall()
     cx.commit()
@@ -127,7 +129,8 @@ def get_url(url_id):
     cu.execute(cmd, (url_id,))
     kws = cu.fetchall()
     kws = map(lambda x: x[0], kws)
-    
+
+    cx.commit()
     return (url, kws)
 
 def update_url(url_id, title, url, abstract, uid, kws):
@@ -205,19 +208,31 @@ def get_last_docs(n):
     if not cx:
 	cx = sqlite3.connect(GINGIN_DB)
     cu = cx.cursor()
-    cmd = 'select doc_id, title, pdate, creator from docs order by pdate desc limit ?'
+    cmd = 'select doc_id, title, pdate, creator from docs where published = 1 order by pdate desc limit ?'
     cu.execute(cmd, (n,))
     docs = cu.fetchall()
     res = map(lambda x: (x[0], x[1], x[2], x[3]), docs)
     cx.commit()
     return res
 
+def get_unpublished_docs():
+    global cx
+    if not cx:
+	cx = sqlite3.connect(GINGIN_DB)
+    cu = cx.cursor()
+    cmd = 'select doc_id, title, pdate, creator from docs where published = 0 order by pdate desc'
+    cu.execute(cmd)
+    docs = cu.fetchall()
+    res = map(lambda x: (x[0], x[1], x[2], x[3]), docs)
+    cx.commit()
+    return res
+
 def get_last_doc_ids(n):
     global cx
     if not cx:
 	cx = sqlite3.connect(GINGIN_DB)
     cu = cx.cursor()
-    cmd = 'select doc_id from docs order by doc_id desc limit ?'
+    cmd = 'select doc_id from docs where published = 1 order by doc_id desc limit ?'
     cu.execute(cmd, (n,))
     docs = cu.fetchall()
     res = map(lambda x: x[0], docs)
@@ -242,7 +257,7 @@ def get_last_abstracts(n):
     if not cx:
 	cx = sqlite3.connect(GINGIN_DB)
     cu = cx.cursor()
-    cmd = 'select doc_id, title, pdate, creator, doc_body from docs order by pdate desc limit ?'
+    cmd = 'select doc_id, title, pdate, creator, doc_body from docs where published = 1 order by pdate desc limit ?'
     cu.execute(cmd, (n,))
     docs = cu.fetchall()
     res = map(lambda x: (x[0], x[1], x[2], x[3], abstract_document(x[4])), docs)
@@ -259,17 +274,21 @@ def get_id_doc(doc_id):
     cmd = 'select * from docs where doc_id = ?'
     cu.execute(cmd, (doc_id,))
     row = cu.fetchone()
+    if not row:
+        raise ValueError, \
+            'The document, with id #%d, is not existed.' % (doc_id)
     title = row[1]
     doc_body = row[2]
     pdate = row[3]
     creator = row[4]
+    published = row[5] and True or False
     # get keywords
     cmd = 'select keyword from keywords a inner join kws_docs b on a.kw_id = b.kw_id where b.doc_id = ?'
     cu.execute(cmd, (doc_id,))
     rows = cu.fetchall()
     kws = map(lambda x: x[0], rows)
     cx.commit()
-    return (kws, doc_id, title, doc_body, pdate, creator)
+    return (kws, doc_id, title, doc_body, pdate, creator, published)
 
 
 def get_id_doc_attack(doc_id, cu=None):
@@ -296,10 +315,11 @@ def get_id_doc_attack(doc_id, cu=None):
 	cmd = 'insert into antispam(doc_id, key, plain, cipher) values(?, ?, ?, ?)'
 	wcu.execute(cmd, (doc_id, key, plain, cipher))
 	attack = key, plain, cipher
-	if cu == None:
-	    cx.commit()
-	    pass
-	pass
+        pass
+    
+    if cu == None:
+        cx.commit()
+        pass
     return attack
 
 
@@ -316,6 +336,9 @@ def get_id_doc_comments(doc_id, cu=None):
     wcu.execute(cmd, (doc_id,))
     rows = wcu.fetchall()
     comments = [(row[0], row[1], row[2], row[3], row[4], row[5]) for row in rows]
+    if cu == None:
+        cx.commit()
+        pass
     return comments
 
 
@@ -428,6 +451,8 @@ def exist_doc_id(doc_id, cu=None):
 	wcu = cx.cursor()
     wcu.execute('select count(*) from docs where doc_id = ?', (doc_id,))
     cnt = wcu.fetchone()[0]
+    if cu == None:
+	cx.commit()
     return cnt > 0
 
 
@@ -438,6 +463,8 @@ def get_afname_mtype_by_afkey(doc_id, afkey, cu=None):
 	if not cx:
 	    cx = sqlite3.connect(GINGIN_DB)
 	wcu = cx.cursor()
+        pass
+
     cmd = 'select afname, mtype from afiles where doc_id = ? and afkey = ?'
     wcu.execute(cmd, (doc_id, afkey))
     rows = wcu.fetchall()
@@ -516,6 +543,27 @@ def remove_afile_by_afkey(doc_id, afkey, cu=None):
 	cx.commit()
     return
 
+def is_doc_published(doc_id, cu=None):
+    global cx
+    wcu = cu
+    if wcu == None:
+	if not cx:
+	    cx = sqlite3.connect(GINGIN_DB)
+	wcu = cx.cursor()
+        pass
+    
+    cmd = 'select 1 from docs where doc_id = ? and published = 1'
+    wcu.execute(cmd, (doc_id,))
+    rows = wcu.fetchall()
+    
+    if cu == None:
+        cx.commit()
+        pass
+
+    if len(rows) != 1:
+        return False
+    return True
+
 
 def get_afkeys_doc(doc_id, cu=None):
     global cx
@@ -524,6 +572,7 @@ def get_afkeys_doc(doc_id, cu=None):
 	if not cx:
 	    cx = sqlite3.connect(GINGIN_DB)
 	wcu = cx.cursor()
+        pass
     
     cmd = 'select afkey from afiles where doc_id = ?'
     wcu.execute(cmd, (doc_id,))
@@ -541,7 +590,7 @@ def get_hot_keywords(hot_n, cu=None):
 	    cx = sqlite3.connect(GINGIN_DB)
 	wcu = cx.cursor()
     
-    cmd = 'select * from hot_kws order by num desc limit ?'
+    cmd = 'select * from hot_kws where num > 0 order by num desc limit ?'
     wcu.execute(cmd, (hot_n,))
     rows = wcu.fetchall()
     kws = map(lambda x: (x[0], x[1]), rows);
@@ -574,6 +623,8 @@ def get_trackback_pings(doc_id, cu=None):
     wcu.execute(cmd, (doc_id,))
     rows = wcu.fetchall()
     pings = map(lambda x: (x[1], x[2], x[3], x[4], x[5], x[6]), rows)
+    if cu == None:
+	cx.commit()
     return pings
 
 def get_referers_of_doc(doc_id, cu=None):
@@ -587,6 +638,8 @@ def get_referers_of_doc(doc_id, cu=None):
     wcu.execute(cmd, (doc_id,))
     rows = wcu.fetchall()
     referers = [x[0] for x in rows]
+    if cu == None:
+	cx.commit()
     return referers
 
 def add_referer(doc_id, url, cu=None):
@@ -602,4 +655,6 @@ def add_referer(doc_id, url, cu=None):
     except sqlite3.DatabaseError:
 	pass
 
+    if cu == None:
+	cx.commit()
     pass
